/*
 * Copyright (C) 2002 Robert Ernst <robert.ernst@linux-solutions.at>
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See COPYING for GPL licensing information.
 *
 */

#include <qpe/qpeapplication.h>
#include <qtimer.h>
#include <qmessagebox.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qkeycode.h>
#include "PlayField.h"
#include <stdlib.h>
#include <stdio.h> // for dump()

PlayField::PlayField(QWidget *parent, const char *name, WFlags f) :
    QWidget(parent, name, f)
{
    m_sprites = new SpriteCollection(this);
    m_timer = new QTimer(this);
    m_pixmap = 0;
    connect(m_timer, SIGNAL(timeout()), this, SLOT(timerTick()));
    m_speed = -1;
    m_strength = -1;
    m_coins = -1;
    m_state = Intro_1;
    m_counter = 0;
    m_score = 0;
    m_lives = 0;
    m_levels_completed = 0;
    m_blocks_broken = 0;
    m_coins_collected = 0;
    m_flames_killed = 0;
    m_actors = 0;
    m_coins_in_level = 0;
    m_click_x = -1;
    m_click_y = -1;
#ifdef CAN_TELEPORT
    m_teleport = false;
#endif
}

PlayField::~PlayField()
{
    delete m_sprites;
    delete m_timer;
    delete m_pixmap;
}

void PlayField::paintEvent(QPaintEvent *event)
{
    switch (m_state) {
    case Intro_1:
	drawIntro_1();
	break;
    case Intro_2:
	drawIntro_2();
	break;
    case Intro_3:
	drawIntro_3();
	break;
    case Playing:
    case Dying:
    case HappyPenguin:
	drawField();
	break;
    case GameOver:
	drawGameOver();
	break;
    }
}

void PlayField::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    m_timer->start(100);
    setFocusPolicy(StrongFocus);
}

void PlayField::hideEvent(QHideEvent *event)
{
    QWidget::hideEvent(event);
    m_timer->stop();
}

void PlayField::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
    if (!m_pixmap || event->size() != event->oldSize()) {
	if (m_pixmap) {
	    delete m_pixmap;
	}
	m_pixmap = new QPixmap(event->size().width(), event->size().height());
    }
}

void PlayField::keyPressEvent(QKeyEvent *event)
{
    QWidget::keyPressEvent(event);
    switch (event->key()) {
    case Key_Up:
	if (m_state == Playing) {
	    m_actor[0].m_req_dir = Up;
	    m_click_x = -1;
	    m_click_y = -1;
#ifdef CAN_TELEPORT
	    m_teleport = false;
#endif
	}
	break;
    case Key_Down:
	if (m_state == Playing) {
	    m_actor[0].m_req_dir = Down;
	    m_click_x = -1;
	    m_click_y = -1;
#ifdef CAN_TELEPORT
	    m_teleport = false;
#endif
	}
	break;
    case Key_Left:
	if (m_state == Playing) {
	    m_actor[0].m_req_dir = Left;
	    m_click_x = -1;
	    m_click_y = -1;
#ifdef CAN_TELEPORT
	    m_teleport = false;
#endif
	}
	break;
    case Key_Right:
	if (m_state == Playing) {
	    m_actor[0].m_req_dir = Right;
	    m_click_x = -1;
	    m_click_y = -1;
#ifdef CAN_TELEPORT
	    m_teleport = false;
#endif
	}
	break;
    case Key_Space:
    case Key_Enter:
	if (m_state != Playing && m_state != Dying && m_state != HappyPenguin) {
	    start();
#ifdef CAN_TELEPORT
	} else if (m_state == Playing) {
	    m_teleport = true;
#endif
	}
	break;
    case Key_Escape:
	if (m_state == Playing || m_state == Dying || m_state == HappyPenguin) {
	    stop();
	}
	break;
    case Key_D:
	dump(__FILE__, __FUNCTION__, __LINE__);
	break;
    }
}

void PlayField::keyReleaseEvent(QKeyEvent *event)
{
    QWidget::keyReleaseEvent(event);
    QWidget::keyPressEvent(event);
    switch (event->key()) {
    case Key_Up:
    case Key_Down:
    case Key_Left:
    case Key_Right:
	if (m_state == Playing) {
	    m_actor[0].m_req_dir = Stop;
	    m_click_x = -1;
	    m_click_y = -1;
#ifdef CAN_TELEPORT
	    m_teleport = false;
#endif
	}
	break;
    }
}

void PlayField::mousePressEvent(QMouseEvent *event)
{
    QWidget::mousePressEvent(event);
    if (m_state == Playing) {
	int max_y = m_pixmap->height();
	int offs_y = (max_y - 240);
#ifdef CAN_TELEPORT
	if (event->y() < offs_y) {
	    m_teleport = true;
	    return;
	} else {
	    m_teleport = false;
	}
#endif
	int xdiff = event->x() - 15 - m_actor[0].m_x;
	int ydiff = event->y() - offs_y - 15 - m_actor[0].m_y;
	if (abs(xdiff) > abs(ydiff)) {
	    ydiff = 0;
	} else {
	    xdiff = 0;
	}
	if (ydiff > 15) {
	    m_actor[0].m_req_dir = Down;
	    m_click_x = (m_actor[0].m_x / 30) * 30;
	    m_click_y = ((event->y() - offs_y) / 30) * 30;
	} else if (ydiff < -15) {
	    m_actor[0].m_req_dir = Up;
	    m_click_x = (m_actor[0].m_x / 30) * 30;
	    m_click_y = ((event->y() - offs_y) / 30) * 30;
	} else if (xdiff > 15) {
	    m_actor[0].m_req_dir = Right;
	    m_click_x = (event->x() / 30) * 30;
	    m_click_y = (m_actor[0].m_y / 30) * 30;
	} else if (xdiff < -15) {
	    m_actor[0].m_req_dir = Left;
	    m_click_x = (event->x() / 30) * 30;
	    m_click_y = (m_actor[0].m_y / 30) * 30;
	} else {
	    m_actor[0].m_req_dir = Stop;
	    m_click_x = -1;
	    m_click_y = -1;
	}
    } else if (m_state != Dying && m_state != HappyPenguin) {
	start();
    }
}

void PlayField::start(void)
{
    m_state = Playing;
    m_counter = 0;
    m_score = 0;
    m_lives = 3;
    m_levels_completed = 0;
    m_blocks_broken = 0;
    m_coins_collected = 0;
    m_flames_killed = 0;
    m_actors = 0;
    m_coins_in_level = 0;
    m_click_x = -1;
    m_click_y = -1;
#ifdef CAN_TELEPORT
    m_teleport = false;
#endif
    int nloops = 0;
    do {
	makeField();
    } while (!isFieldOk() && ++nloops < 10);
    if (nloops >= 10) {
	dump(__FILE__, __FUNCTION__, __LINE__);
    }
    repaint(false);
}

void PlayField::stop(void)
{
    if (m_state != Intro_1) {
	m_state = Intro_1;
	m_counter = 0;
	repaint(false);
    }
}

void PlayField::timerTick(void)
{
    m_counter++;

    switch (m_state) {
    case Intro_1:
	if (m_counter >= 100) {
	    m_state = Intro_2;
	    m_counter = 0;
	}
	break;
    case Intro_2:
	if (m_counter >= 100) {
	    m_state = Intro_3;
	    m_counter = 0;
	}
	break;
    case Intro_3:
	if (m_counter >= 100) {
	    m_state = Intro_1;
	    m_counter = 0;
	}
	break;
    case Playing:
	/* clear the actors that can move from the drawing buffer */
	clearField();
	/* update the counters for actors that have an animated pixmap */
	animateActors();
	/* update the actors positions */
	moveActors();
	/* re-create flames from points after a few seconds */
	recreateFlames();
	break;
    case Dying:
	/* clear the actors that can move from the drawing buffer */
	clearField();
	/* update the counters for actors that have an animated pixmap */
	animateActors();
	/* update the actors positions */
	recreateFlames();
	/* leave this state after a while */
	if (m_actor[0].m_counter >= 32) {
	    if (m_lives > 0) {
		m_state = Playing;
		recreatePlayer();
	    } else {
		m_state = GameOver;
		m_counter = 0;
	    }
	}
	break;
    case HappyPenguin:
	/* clear the actors that can move from the drawing buffer */
	clearField();
	/* update the counters for actors that have an animated pixmap */
	animateActors();
	/* re-create flames from points after a few seconds */
	recreateFlames();
	/* leave this state after a while */
	if (m_actor[0].m_counter >= 40) {
	    m_actors = 0;
	    m_coins_in_level = 0;
	    int nloops = 0;
	    do {
		makeField();
	    } while (!isFieldOk() && ++nloops < 10);
	    if (nloops >= 10) {
		dump(__FILE__, __FUNCTION__, __LINE__);
	    }
	    m_state = Playing;
	    m_actor[0].m_counter = 0;
	    m_actor[0].m_req_dir = Stop;
	    m_actor[0].m_act_dir = Stop;
	    m_click_x = -1;
	    m_click_y = -1;
#ifdef CAN_TELEPORT
	    m_teleport = false;
#endif
	}
	break;
    case GameOver:
	if (m_counter >= 300) {
	    m_state = Intro_1;
	    m_counter = 0;
	}
	break;
    }

    repaint(false);
}

void PlayField::updateSpeed(int speed)
{
    if (speed != m_speed) {
	m_speed = speed;
	if (m_state == Playing || m_state == Dying || m_state == HappyPenguin) {
	    start();
	}
    }
}

void PlayField::updateStrength(int strength)
{
    if (strength != m_strength) {
	m_strength = strength;
	if (m_state == Playing || m_state == Dying || m_state == HappyPenguin) {
	    start();
	}
    }
}

void PlayField::updateCoins(int coins)
{
    if (coins != m_coins) {
	m_coins = coins;
	if (m_state == Playing || m_state == Dying || m_state == HappyPenguin) {
	    start();
	}
    }
}

void PlayField::drawIntro_1(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int max_x = width();
    int max_y = height();

    painter_pixmap.fillRect(0, 0, max_x, max_y, black);
    painter_pixmap.setPen(white);

    m_sprites->drawIntro(painter_pixmap, (max_x - 224) / 2, 7);
    painter_pixmap.drawText(0, 68, max_x, 14, AlignHCenter | AlignBottom,
	tr("Actors and Objects"));
    m_sprites->drawHappyPenguin(painter_pixmap, 15, 90, (m_counter % 10) / 5);
    switch ((m_counter % 36) / 9) {
    case 0:
	m_sprites->drawPenguinDown(painter_pixmap, 45, 90, (m_counter % 12) / 3);
	break;
    case 1:
	m_sprites->drawPenguinRight(painter_pixmap, 45, 90, (m_counter % 12) / 3);
	break;
    case 2:
	m_sprites->drawPenguinUp(painter_pixmap, 45, 90, (m_counter % 12) / 3);
	break;
    case 3:
	m_sprites->drawPenguinLeft(painter_pixmap, 45, 90, (m_counter % 12) / 3);
	break;
    }
    painter_pixmap.drawText(100, 90, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Pixel Pete, the penguin"));
    m_sprites->drawFlame(painter_pixmap, 15, 124, 0);
    m_sprites->drawFlame(painter_pixmap, 45, 124, (m_counter % 16) / 2);
    painter_pixmap.drawText(100, 124, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Evil flames, they kill"));
    m_sprites->drawBlock(painter_pixmap, 15, 158, 0);
    m_sprites->drawBlock(painter_pixmap, 45, 158, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 158, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Ice cubes - push them"));
    m_sprites->drawBlockCoin(painter_pixmap, 15, 192, 0);
    m_sprites->drawBlockCoin(painter_pixmap, 45, 192, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 192, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Frozen coins - get them all"));
    m_sprites->drawWall(painter_pixmap, 15, 226);
    painter_pixmap.drawText(100, 226, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Solid walls"));

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void PlayField::drawIntro_2(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int max_x = width();
    int max_y = height();

    painter_pixmap.fillRect(0, 0, max_x, max_y, black);
    painter_pixmap.setPen(white);

    m_sprites->drawIntro(painter_pixmap, (max_x - 224) / 2, 7);
    painter_pixmap.drawText(0, 68, max_x, 14, AlignHCenter | AlignBottom,
	tr("How to play the game"));
    m_sprites->drawPenguinDown(painter_pixmap, 15, 90, (m_counter % 12) / 3);
    painter_pixmap.drawText(100, 90, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Move up, down, left, right"));
    painter_pixmap.drawText(100, 105, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("with the directional keys"));
    m_sprites->drawPenguinRight(painter_pixmap, 15, 124, (m_counter % 12) / 3);
    m_sprites->drawBlock(painter_pixmap, 45, 124, 0);
    painter_pixmap.drawText(100, 124, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Walk against ice cubes to"));
    painter_pixmap.drawText(100, 139, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("move them out of the way"));
    m_sprites->drawPenguinRight(painter_pixmap, 15, 158, (m_counter % 12) / 3);
    m_sprites->drawBlock(painter_pixmap, 45, 158, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 158, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Walk against blocked ice"));
    painter_pixmap.drawText(100, 173, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("cubes to crack them"));
    m_sprites->drawPenguinRight(painter_pixmap, 15, 192, (m_counter % 12) / 3);
    m_sprites->drawBlockCoin(painter_pixmap, 45, 192, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 192, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Free the gold coins by"));
    painter_pixmap.drawText(100, 207, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("crushing the ice around them"));
    m_sprites->drawDead(painter_pixmap, 15, 226, (m_counter % 16) / 2);
    m_sprites->drawFlame(painter_pixmap, 45, 226, (m_counter % 16) / 2);
    painter_pixmap.drawText(100, 226, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("And watch out"));
    painter_pixmap.drawText(100, 241, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("for the flames"));

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void PlayField::drawIntro_3(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int max_x = width();
    int max_y = height();

    painter_pixmap.fillRect(0, 0, max_x, max_y, black);
    painter_pixmap.setPen(white);

    m_sprites->drawIntro(painter_pixmap, (max_x - 224) / 2, 7);
    painter_pixmap.drawText(0, 68, max_x, 14, AlignHCenter | AlignBottom,
	tr("How to score points"));
    m_sprites->drawPenguinRight(painter_pixmap, 15, 90, (m_counter % 12) / 3);
    m_sprites->drawBlock(painter_pixmap, 45, 90, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 90, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Breaking ice:"));
    painter_pixmap.drawText(100, 105, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("5 points"));
    m_sprites->drawFlame(painter_pixmap, 15, 124, (m_counter % 16) / 2);
    painter_pixmap.drawText(100, 124, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Putting out flame with ice:"));
    painter_pixmap.drawText(100, 139, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("50 points"));
    m_sprites->drawPenguinRight(painter_pixmap, 15, 158, (m_counter % 12) / 3);
    m_sprites->drawBlockCoin(painter_pixmap, 45, 158, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 158, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Freeing coin:"));
    painter_pixmap.drawText(100, 173, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("100 points"));
    m_sprites->drawCoin(painter_pixmap, 15, 192);
    painter_pixmap.drawText(100, 192, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("Taking all coins:"));
    painter_pixmap.drawText(100, 207, max_x - 100, 15, AlignLeft | AlignVCenter,
	tr("1000 points"));
    painter_pixmap.drawText(0, 226, max_x, max_y - 226, AlignHCenter | AlignVCenter,
	tr("Press space to start game"));

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void PlayField::drawField(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int max_x = m_pixmap->width();
    int max_y = m_pixmap->height();
    int offs_y = (max_y - 240);
    int offs = (max_y - 243 - 15) / 2;
    int i;

    painter_pixmap.fillRect(0, 0, offs + max_lives * 15, max_y - 243, black);
    painter_pixmap.setPen(white);
    for (i = 0; i < m_lives; i++) {
	m_sprites->drawLive(painter_pixmap, offs + 15 * i, offs);
    }
    painter_pixmap.fillRect(max_x - 90 - offs, offs, 15, 15, black);
    painter_pixmap.drawText(max_x - 90 - offs, offs, 15, 15, AlignRight | AlignVCenter,
	QString("%1").arg(m_levels_completed + 1));
    painter_pixmap.fillRect(max_x - 35 - offs, offs, 35, 15, black);
    painter_pixmap.drawText(max_x - 35 - offs, offs, 35, 15, AlignRight | AlignVCenter,
	QString("%1").arg(m_score));

    for (i = 0; i < m_actors; i++) {
	switch (m_actor[i].m_type) {
	case Player:
	    if (m_state == Dying) {
		m_sprites->drawDead(painter_pixmap, m_actor[i].m_x, m_actor[i].m_y + offs_y,
		    m_actor[i].m_counter < 16 ? (m_actor[i].m_counter % 16) / 2 : 7);
	    } else if (m_state == HappyPenguin) {
		m_sprites->drawHappyPenguin(painter_pixmap, m_actor[i].m_x,
		    m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 10) / 5);
	    } else if (m_actor[i].m_act_dir == Up) {
		m_sprites->drawPenguinUp(painter_pixmap, m_actor[i].m_x,
		    m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 12) / 3);
	    } else if (m_actor[i].m_act_dir == Down) {
		m_sprites->drawPenguinDown(painter_pixmap, m_actor[i].m_x,
		    m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 12) / 3);
	    } else if (m_actor[i].m_act_dir == Left) {
		m_sprites->drawPenguinLeft(painter_pixmap, m_actor[i].m_x,
		    m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 12) / 3);
	    } else if (m_actor[i].m_act_dir == Right) {
		m_sprites->drawPenguinRight(painter_pixmap, m_actor[i].m_x,
		    m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 12) / 3);
	    } else {
		m_sprites->drawPenguin(painter_pixmap, m_actor[i].m_x,
		    m_actor[i].m_y + offs_y);
	    }
	    break;
	case Block:
	    m_sprites->drawBlock(painter_pixmap, m_actor[i].m_x,
		m_actor[i].m_y + offs_y, m_actor[i].m_counter);
	    break;
	case BlockCoin:
	    m_sprites->drawBlockCoin(painter_pixmap, m_actor[i].m_x,
		m_actor[i].m_y + offs_y, m_actor[i].m_counter);
	    break;
	case Flame:
	    m_sprites->drawFlame(painter_pixmap, m_actor[i].m_x,
		m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 16) / 2);
	    break;
	case Points:
	    m_sprites->drawPoints(painter_pixmap, m_actor[i].m_x,
		m_actor[i].m_y + offs_y, (m_actor[i].m_counter % 6) / 3);
	    break;
	case Wall:
	    m_sprites->drawWall(painter_pixmap, m_actor[i].m_x,
		m_actor[i].m_y + offs_y);
	    break;
	default:
	    break;
	}
    }

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void PlayField::drawGameOver(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int max_x = width();
    int max_y = height();

    painter_pixmap.fillRect(0, 0, max_x, max_y, black);
    painter_pixmap.setPen(white);

    m_sprites->drawIntro(painter_pixmap, (max_x - 224) / 2, 7);
    painter_pixmap.drawText(0, 68, max_x, 14, AlignHCenter | AlignBottom,
	tr("Game over"));
    m_sprites->drawHappyPenguin(painter_pixmap, 15, 90, (m_counter % 10) / 5);
    painter_pixmap.drawText(100, 90, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Levels completed: %1").arg(m_levels_completed));
    m_sprites->drawBlock(painter_pixmap, 15, 124, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 124, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Blocks broken: %1").arg(m_blocks_broken));
    m_sprites->drawBlockCoin(painter_pixmap, 15, 158, (m_counter % 32) / 4);
    painter_pixmap.drawText(100, 158, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Coins collected: %1").arg(m_coins_collected));
    m_sprites->drawFlame(painter_pixmap, 15, 192, (m_counter % 16) / 2);
    painter_pixmap.drawText(100, 192, max_x - 100, 30, AlignLeft | AlignVCenter,
	tr("Flames killed: %1").arg(m_flames_killed));
    painter_pixmap.drawText(0, 226, max_x, max_y - 226, AlignHCenter | AlignVCenter,
	tr("Score: %1 %2").arg(m_score).arg(m_score == 1 ? tr("point") : tr("points")));

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void PlayField::makeField(void)
{
    QPainter painter_pixmap(m_pixmap);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int max_x = m_pixmap->width();
    int max_y = m_pixmap->height();
    int offs_y = (max_y - 240);
    int offs = (max_y - 243 - 15) / 2;
    int i;
    int j;

    painter_pixmap.fillRect(0, 0, max_x, max_y, black);
    painter_pixmap.setPen(white);
    painter_pixmap.drawLine(0, max_y - 241, max_x, max_y - 241);
    for (i = 0; i < m_lives; i++) {
	m_sprites->drawLive(painter_pixmap, offs + 15 * i, offs);
    }
    painter_pixmap.drawText(max_x - 120 - offs, offs, 30, 15, AlignLeft | AlignVCenter,
	tr("Level:"));
    painter_pixmap.drawText(max_x - 90 - offs, offs, 15, 15, AlignRight | AlignVCenter,
	QString("%1").arg(m_levels_completed + 1));
    painter_pixmap.drawText(max_x - 70 - offs, offs, 35, 15, AlignLeft | AlignVCenter,
	tr("Score:"));
    painter_pixmap.drawText(max_x - 35 - offs, offs, 35, 15, AlignRight | AlignVCenter,
	QString("%1").arg(m_score));

    /* reset all actors */
    memset(m_actor, 0, sizeof (m_actor));
    m_actors = 0;
    m_coins_in_level = 0;

    /* pixel pete, the penguin */
    m_actor[0].m_speed = m_speed;
    m_actor[0].m_type = Player;
    m_actors++;

    /* the ice blocks with and without coins */
    int level_blocks[max_config] = {17, 16, 15, 14, 13, 12};
    for (i = 0; i < level_blocks[m_levels_completed % max_config]; i++) {
	int nloops = 0;
	while (++nloops < 1000) {
	    int x = (rand() % max_cols) * 30;
	    int y = (rand() % max_rows) * 30;
	    for (j = 0; j < m_actors; j++) {
		if (m_actor[j].m_x == x && m_actor[j].m_y == y) {
		    break;
		}
	    }
	    if (j >= m_actors) {
		m_actor[m_actors].m_x = x; 
		m_actor[m_actors].m_y = y;
		m_actor[m_actors].m_speed = m_strength; // they move when pushed...
		m_actor[m_actors].m_type = Block;
		m_actors++;
		break;
	    }
	}
	if (nloops >= 1000) {
	    dump(__FILE__, __FUNCTION__, __LINE__);
	}
    }

    /* the evil flames */
    int level_flames[max_config] = {2, 3, 4, 2, 3, 4};
    int level_flames_speed[max_config] = {3, 3, 3, 5, 5, 5};
    for (i = 0; i < level_flames[m_levels_completed % max_config]; i++) {
	int nloops = 0;
	while (++nloops < 1000) {
	    int x = (rand() % max_cols) * 30;
	    int y = (rand() % max_rows) * 30;
	    if (x < (max_cols * 15) && y < (max_rows * 15)) {
		/* move the flame to make the distance to the penguin larger */
		x += max_cols * 15;
		y += max_rows * 15;
	    }
	    for (j = 0; j < m_actors; j++) {
		if (m_actor[j].m_x == x && m_actor[j].m_y == y) {
		    break;
		}
	    }
	    if (j >= m_actors) {
		m_actor[m_actors].m_x = x;
		m_actor[m_actors].m_y = y;
		m_actor[m_actors].m_speed
		    = level_flames_speed[m_levels_completed % max_config];
		m_actor[m_actors].m_type = Flame;
		m_actors++;
		break;
	    }
	}
	if (nloops >= 1000) {
	    dump(__FILE__, __FUNCTION__, __LINE__);
	}
    }

    /* the walls */
    int level_walls[max_config] = {5, 6, 7, 8, 9, 10};
    for (i = 0; i < level_walls[m_levels_completed % max_config]; i++) {
	int nloops = 0;
	while (++nloops < 1000) {
	    int x = (rand() % max_cols) * 30;
	    int y = (rand() % max_rows) * 30;
	    for (j = 0; j < m_actors; j++) {
		if (m_actor[j].m_x == x && m_actor[j].m_y == y) {
		    break;
		}
	    }
	    if (j >= m_actors) {
		m_actor[m_actors].m_x = x;
		m_actor[m_actors].m_y = y;
		m_actor[m_actors].m_type = Wall;
		m_actors++;
		/* draw the wall now, it will never move */
		m_sprites->drawWall(painter_pixmap, x, y + offs_y);
		break;
	    }
	}
	if (nloops >= 1000) {
	    dump(__FILE__, __FUNCTION__, __LINE__);
	}
    }

    /* the coins, always frozen inside an ice block */
    for (i = 0; i < m_coins; i++) {
	int nloops = 0;
	while (++nloops < 1000) {
	    int x = (rand() % max_cols) * 30;
	    int y = (rand() % max_rows) * 30;
	    for (j = 0; j < m_actors; j++) {
		if (m_actor[j].m_x == x && m_actor[j].m_y == y
		    && m_actor[j].m_type == Block) {
		    break;
		}
	    }
	    if (j < m_actors) {
		m_actor[j].m_type = BlockCoin;
		break;
	    }
	}
	if (nloops >= 1000) {
	    dump(__FILE__, __FUNCTION__, __LINE__);
	}
    }
    m_coins_in_level = m_coins;
}

void PlayField::check_field(enum Types map[max_rows][max_cols],
    bool visited[max_rows][max_cols], int row, int col) {
    visited[row][col] = true;
    if (row > 0 && map[row - 1][col] != Wall
	&& visited[row - 1][col] == false) {
	check_field(map, visited, row - 1, col);
    }
    if (row < (max_rows - 1) && map[row + 1][col] != Wall
	&& visited[row + 1][col] == false) {
	check_field(map, visited, row + 1, col);
    }
    if (col > 0 && map[row][col - 1] != Wall
	&& visited[row][col - 1] == false) {
	check_field(map, visited, row, col - 1);
    }
    if (col < (max_cols - 1) && map[row][col + 1] != Wall
	&& visited[row][col + 1] == false) {
	check_field(map, visited, row, col + 1);
    }
};
bool PlayField::isFieldOk(void)
{
    enum Types map[max_rows][max_cols];
    bool visited[max_rows][max_cols];
    int i;
    int j;

    /* first, set all fields to "unvisited" and build the map */
    for (i = 0; i < max_rows; i++) {
	for (j = 0; j < max_cols; j++) {
	    visited[i][j] = false;
	    map[i][j] = Empty;
	}
    }
    for (i = 0; i < m_actors; i++) {
	int x = m_actor[i].m_x / 30;
	int y = m_actor[i].m_y / 30;
	if (x < 0 || x >= max_cols || y < 0 || y >= max_rows) {
	    dump(__FILE__, __FUNCTION__, __LINE__);
	    return false;
	}
	map[y][x] = m_actor[i].m_type;
    }

    /* second, set all field to "visited" that can be reached from x=0/y=0 */
    check_field(map, visited, 0, 0);

    /* third, check if all coins are on a visited field */
    for (i = 0; i < max_rows; i++) {
	for (j = 0; j < max_cols; j++) {
	    if (map[i][j] == BlockCoin && visited[i][j] == false) {
		return false;
	    }
	}
    }
    return true;
}

void PlayField::clearField(void)
{
    QPainter painter_pixmap(m_pixmap);
    QColor black(0, 0, 0);
    int max_y = m_pixmap->height();
    int offs_y = (max_y - 240);
    int i;

    for (i = 0; i < m_actors; i++) {
	if (m_actor[i].m_type != Wall) {
	    painter_pixmap.fillRect(m_actor[i].m_x, m_actor[i].m_y + offs_y,
		30, 30, black);
	}
    }
}

void PlayField::animateActors(void)
{
    int i;

    for (i = 0; i < m_actors; i++) {
	if (m_actor[i].m_type == Player || m_actor[i].m_type == Flame
	    || m_actor[i].m_type == Points) {
	    m_actor[i].m_counter++;
	}
    }
}

void PlayField::moveActors(void)
{
    int x;
    int y;
    int i;
    int j;

    /* first, move the blocks - they just move until they crash
     * into another object, but they kill all flames on their way...
     */
    for (i = 1; i < m_actors; i++) {
	if (m_actor[i].m_type == Block || m_actor[i].m_type == BlockCoin) {
	    switch (m_actor[i].m_act_dir) {
	    case Up:
		x = m_actor[i].m_x;
		y = m_actor[i].m_y - m_actor[i].m_speed;
		while ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killFlame(j);
		}
		if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    y = (m_actor[i].m_y / 30) * 30;
		    m_actor[i].m_act_dir = Stop;
		    m_actor[i].m_req_dir = Stop;
		}
		m_actor[i].m_x = x;
		m_actor[i].m_y = y;
		break;
	    case Down:
		x = m_actor[i].m_x;
		y = m_actor[i].m_y + m_actor[i].m_speed;
		while ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killFlame(j);
		}
		if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    y = ((m_actor[i].m_y + 29) / 30) * 30;
		    m_actor[i].m_act_dir = Stop;
		    m_actor[i].m_req_dir = Stop;
		}
		m_actor[i].m_x = x;
		m_actor[i].m_y = y;
		break;
	    case Left:
		x = m_actor[i].m_x - m_actor[i].m_speed;
		y = m_actor[i].m_y;
		while ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killFlame(j);
		}
		if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    x = (m_actor[i].m_x / 30) * 30;
		    m_actor[i].m_act_dir = Stop;
		    m_actor[i].m_req_dir = Stop;
		}
		m_actor[i].m_x = x;
		m_actor[i].m_y = y;
		break;
	    case Right:
		x = m_actor[i].m_x + m_actor[i].m_speed;
		y = m_actor[i].m_y;
		while ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killFlame(j);
		}
		if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    x = ((m_actor[i].m_x + 29) / 30) * 30;
		    m_actor[i].m_act_dir = Stop;
		    m_actor[i].m_req_dir = Stop;
		}
		m_actor[i].m_x = x;
		m_actor[i].m_y = y;
		break;
	    default:
		break;
	    }
	}
    }

    /* then, calculate the directions of the flames (always towards the player) */
    for (i = 1; i < m_actors; i++) {
	if (m_actor[i].m_type == Flame) {
	    /* default: move in a random direction */
	    int xdiff = m_actor[i].m_x - m_actor[0].m_x;
	    int ydiff = m_actor[i].m_y - m_actor[0].m_y;
	    switch (rand() % 5) {
	    case 0:
		m_actor[i].m_req_dir = Stop;
		break;
	    case 1:
		m_actor[i].m_req_dir = Up;
		break;
	    case 2:
		m_actor[i].m_req_dir = Down;
		break;
	    case 3:
		m_actor[i].m_req_dir = Left;
		break;
	    case 4:
		m_actor[i].m_req_dir = Right;
		break;
	    }

	    /* if the player is too far away, the flame doesn't see him */
	    if (abs(xdiff) > 150 || abs(ydiff) > 150) {
		continue;
	    }

	    /* if the flame is in line of sight, it chases the player */
	    if (abs(xdiff) < 15) {
		if (lineOfSight(i, 0, (ydiff > 0) ? Up : Down)) {
		    m_actor[i].m_req_dir = (ydiff > 0) ? Up : Down;
		    continue;
		}
	    } else if (abs(ydiff) < 15) {
		if (lineOfSight(i, 0, (xdiff > 0) ? Left : Right)) {
		    m_actor[i].m_req_dir = (xdiff > 0) ? Left : Right;
		    continue;
		}
	    }

	    /* if the player is too far away, the flame doesn't see him */
	    if (abs(xdiff) > 90 || abs(ydiff) > 90) {
		continue;
	    }

	    /* move in the direction with the smaller difference if we have
	     * to go both horizontal or vertical
	     */
	    if (xdiff != 0 && ydiff != 0) {
		if (abs(xdiff) > abs(ydiff)) {
		    xdiff = 0;
		} else {
		    ydiff = 0;
		}
	    }
	    if (xdiff != 0) {
		m_actor[i].m_req_dir = (xdiff > 0) ? Left : Right;
	    } else if (ydiff != 0) {
		m_actor[i].m_req_dir = (ydiff > 0) ? Up : Down;
	    }
	}
    }

    /* at last, move the player and the flames. note: if the requested direction
     * differs from the actual one, the player or flame is moved to the next tile
     * position before actually changing the direction (or stopping).
     */
    for (i = 0; i < m_actors; i++) {
	if (m_actor[i].m_type != Player && m_actor[i].m_type != Flame) {
	    continue;
	}
#ifdef CAN_TELEPORT
	if (i == 0 && m_teleport) {
	    /* if teleport() works, the act/req_dir members will be reset! */
	    teleport();
	    m_teleport = false;
	}
#endif
	if (m_actor[i].m_act_dir == Stop) {
	    m_actor[i].m_act_dir = m_actor[i].m_req_dir;
	}
	switch (m_actor[i].m_act_dir) {
	case Up:
	    x = m_actor[i].m_x;
	    y = m_actor[i].m_y - m_actor[i].m_speed;
	    if (m_actor[i].m_req_dir != Up) {
		if ((m_actor[i].m_y % 30) < m_actor[i].m_speed) {
		    m_actor[i].m_act_dir = m_actor[i].m_req_dir;
		    y = (m_actor[i].m_y / 30) * 30;
		}
	    } else if (i == 0) {
		if ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    if (m_actor[j].m_type == Block || m_actor[j].m_type == BlockCoin) {
			moveOrCrushBlock(j, m_actor[i].m_act_dir);
		    } else {
			m_actor[i].m_req_dir = Stop;
			m_actor[i].m_act_dir = Stop;
		    }
		    y = (m_actor[i].m_y / 30) * 30;
		}
	    } else {
		if (( j = detectCollisionPlayer(x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionMovingBlock(i, x, y)) >= 0) {
		    killFlame(i);
		    y = (m_actor[i].m_y / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    y = (m_actor[i].m_y / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		}
	    }
	    if (i == 0 && hasReachedClickPosition(x, y)) {
		m_actor[i].m_req_dir = Stop;
	    }
	    m_actor[i].m_x = x;
	    m_actor[i].m_y = y;
	    break;
	case Down:
	    x = m_actor[i].m_x;
	    y = m_actor[i].m_y + m_actor[i].m_speed;
	    if (i == 0 && hasReachedClickPosition(x, y)) {
		m_actor[i].m_req_dir = Stop;
	    }
	    if (m_actor[i].m_req_dir != Down) {
		if ((y % 30) <= m_actor[i].m_speed) {
		    m_actor[i].m_act_dir = m_actor[i].m_req_dir;
		    y = (y / 30) * 30;
		}
	    } else if (i == 0) {
		if ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    if (m_actor[j].m_type == Block || m_actor[j].m_type == BlockCoin) {
			moveOrCrushBlock(j, m_actor[i].m_act_dir);
		    } else {
			m_actor[i].m_req_dir = Stop;
			m_actor[i].m_act_dir = Stop;
		    }
		    y = ((m_actor[i].m_y + 29) / 30) * 30;
		}
	    } else {
		if (( j = detectCollisionPlayer(x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionMovingBlock(i, x, y)) >= 0) {
		    killFlame(i);
		    y = ((m_actor[i].m_y + 29) / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    y = ((m_actor[i].m_y + 29) / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		}
	    }
	    m_actor[i].m_x = x;
	    m_actor[i].m_y = y;
	    break;
	case Left:
	    x = m_actor[i].m_x - m_actor[i].m_speed;
	    y = m_actor[i].m_y;
	    if (m_actor[i].m_req_dir != Left) {
		if ((m_actor[i].m_x % 30) < m_actor[i].m_speed) {
		    m_actor[i].m_act_dir = m_actor[i].m_req_dir;
		    x = (m_actor[i].m_x / 30) * 30;
		}
	    } else if (i == 0) {
		if ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    if (m_actor[j].m_type == Block || m_actor[j].m_type == BlockCoin) {
			moveOrCrushBlock(j, m_actor[i].m_act_dir);
		    } else {
			m_actor[i].m_req_dir = Stop;
			m_actor[i].m_act_dir = Stop;
		    }
		    x = (m_actor[i].m_x / 30) * 30;
		}
	    } else {
		if (( j = detectCollisionPlayer(x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionMovingBlock(i, x, y)) >= 0) {
		    killFlame(i);
		    x = (m_actor[i].m_x / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    x = (m_actor[i].m_x / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		}
	    }
	    if (i == 0 && hasReachedClickPosition(x, y)) {
		m_actor[i].m_req_dir = Stop;
	    }
	    m_actor[i].m_x = x;
	    m_actor[i].m_y = y;
	    break;
	case Right:
	    x = m_actor[i].m_x + m_actor[i].m_speed;
	    y = m_actor[i].m_y;
	    if (i == 0 && hasReachedClickPosition(x, y)) {
		m_actor[i].m_req_dir = Stop;
	    }
	    if (m_actor[i].m_req_dir != Right) {
		if ((x % 30) <= m_actor[i].m_speed) {
		    m_actor[i].m_act_dir = m_actor[i].m_req_dir;
		    x = (x / 30) * 30;
		}
	    } else if (i == 0) {
		if ((j = detectCollisionFlame(i, x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    if (m_actor[j].m_type == Block || m_actor[j].m_type == BlockCoin) {
			moveOrCrushBlock(j, m_actor[i].m_act_dir);
		    } else {
			m_actor[i].m_req_dir = Stop;
			m_actor[i].m_act_dir = Stop;
		    }
		    x = ((m_actor[i].m_x + 29) / 30) * 30;
		}
	    } else {
		if (( j = detectCollisionPlayer(x, y)) >= 0) {
		    killPlayer();
		} else if ((j = detectCollisionMovingBlock(i, x, y)) >= 0) {
		    killFlame(i);
		    x = ((m_actor[i].m_x + 29) / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		} else if ((j = detectCollisionBlockOrWall(i, x, y)) >= 0) {
		    x = ((m_actor[i].m_x + 29) / 30) * 30;
		    m_actor[i].m_req_dir = Stop;
		    m_actor[i].m_act_dir = Stop;
		}
	    }
	    m_actor[i].m_x = x;
	    m_actor[i].m_y = y;
	    break;
	case Stop:
	    break;
	}
    }
}

void PlayField::recreateFlames(void)
{
    int i;

    for (i = 1; i < m_actors; i++) {
	if (m_actor[i].m_type == Points && m_actor[i].m_counter >= 36) {
	    /* first, set the position of the flame to the tile position */
	    m_actor[i].m_x = ((m_actor[i].m_x + 15) / 30) * 30;
	    m_actor[i].m_y = ((m_actor[i].m_y + 15) / 30) * 30;
	    m_actor[i].m_counter = 0;
	    m_actor[i].m_type = Flame;

	    /* then, move the flame if it is too near to the player */
	    int nloops = 0;
	    while (++nloops < 1000) {
		int xdiff = abs(m_actor[i].m_x - m_actor[0].m_x);
		int ydiff = abs(m_actor[i].m_y - m_actor[0].m_y);
		if (xdiff >= 90 && ydiff >= 90
		    && detectCollision(i, m_actor[i].m_x, m_actor[i].m_y) < 0) {
		    break;
		}
		m_actor[i].m_x = (rand() % max_cols) * 30;
		m_actor[i].m_y = (rand() % max_rows) * 30;
	    }
	    if (nloops >= 1000) {
		dump(__FILE__, __FUNCTION__, __LINE__);
	    }
	}
    }
}

void PlayField::recreatePlayer(void)
{
    int i;

    /* first, set the position of the player to the tile position */
    m_actor[0].m_x = ((m_actor[0].m_x + 15) / 30) * 30;
    m_actor[0].m_y = ((m_actor[0].m_y + 15) / 30) * 30;
    m_actor[0].m_counter = 0;

    /* actually, we don't recreate or move the player, but we move all flames
     * that are too near to an other position.
     */
    for (i = 1; i < m_actors; i++) {
	if (m_actor[i].m_type == Flame) {
	    int nloops = 0;
	    while (++nloops < 1000) {
		int xdiff = abs(m_actor[i].m_x - m_actor[0].m_x);
		int ydiff = abs(m_actor[i].m_y - m_actor[0].m_y);
		if (xdiff >= 90 && ydiff >= 90
		    && detectCollision(i, m_actor[i].m_x, m_actor[i].m_y) < 0) {
		    break;
		}
		m_actor[i].m_x = (rand() % max_cols) * 30;
		m_actor[i].m_y = (rand() % max_rows) * 30;
		m_actor[i].m_counter = 0;
	    }
	    if (nloops >= 1000) {
		dump(__FILE__, __FUNCTION__, __LINE__);
	    }
	}
    }
}

void PlayField::killFlame(int actor)
{
    if (actor < 0 || actor >= m_actors || m_actor[actor].m_type != Flame) {
	dump(__FILE__, __FUNCTION__, __LINE__);
    }
    m_flames_killed++;
    m_score = m_score + 50;
    if (m_flames_killed % 10 == 0 && m_lives < max_lives) {
	m_lives++;
    }
    m_actor[actor].m_type = Points;
    m_actor[actor].m_counter = 0;
    m_actor[actor].m_req_dir = Stop;
    m_actor[actor].m_act_dir = Stop;
}

void PlayField::killPlayer(void)
{
    if (m_lives) {
	m_lives--;
    }
    m_state = Dying;
    m_actor[0].m_counter = 0;
    m_actor[0].m_req_dir = Stop;
    m_actor[0].m_act_dir = Stop;
    m_click_x = -1;
    m_click_y = -1;
#ifdef CAN_TELEPORT
    m_teleport = false;
#endif
}

void PlayField::moveOrCrushBlock(int actor, enum Directions dir)
{
    bool moveable = true;
    int j;

    if (actor < 0 || actor >= m_actors
	|| (m_actor[actor].m_type != Block && m_actor[actor].m_type != BlockCoin)) {
	dump(__FILE__, __FUNCTION__, __LINE__);
    }

    /* determine if the block is moveable */
    switch (dir) {
    case Up:
	if (m_actor[actor].m_y < 30) {
	    moveable = false;
	    break;
	}
	j = detectCollisionBlockOrWall(actor, m_actor[actor].m_x, m_actor[actor].m_y - 1);
	if (j >= 0) {
	    moveable = false;
	}
	break;
    case Down:
	if (m_actor[actor].m_y >= ((max_rows - 1) * 30)) {
	    moveable = false;
	    break;
	}
	j = detectCollisionBlockOrWall(actor, m_actor[actor].m_x, m_actor[actor].m_y + 1);
	if (j >= 0) {
	    moveable = false;
	}
	break;
    case Left:
	if (m_actor[actor].m_x < 30) {
	    moveable = false;
	    break;
	}
	j = detectCollisionBlockOrWall(actor, m_actor[actor].m_x - 1, m_actor[actor].m_y);
	if (j >= 0) {
	    moveable = false;
	}
	break;
    case Right:
	if (m_actor[actor].m_x >= ((max_cols - 1) * 30)) {
	    moveable = false;
	    break;
	}
	j = detectCollisionBlockOrWall(actor, m_actor[actor].m_x + 1, m_actor[actor].m_y);
	if (j >= 0) {
	    moveable = false;
	}
	break;
    default:
	break;
    }

    /* now, move or crush the block */
    if (moveable) {
	m_actor[actor].m_req_dir = dir;
	m_actor[actor].m_act_dir = dir;
    } else {
	m_actor[actor].m_counter = m_actor[actor].m_counter
	    + m_strength / 6;
	if (m_actor[actor].m_counter >= 8) {
	    if (m_actor[actor].m_type == BlockCoin) {
		m_coins_collected++;
		m_score = m_score + 100;
		m_coins_in_level--;
		if (m_coins_in_level == 0) {
		    m_levels_completed++;
		    m_score = m_score + 1000;
		    m_state = HappyPenguin;
		    m_actor[0].m_counter = 0;
		    if (m_levels_completed % 10 == 0 && m_lives < max_lives) {
			m_lives++;
		    }
		}
	    } else {
		m_blocks_broken++;
		m_score = m_score + 5;
	    }
	    m_actor[actor].m_type = Empty;
	}
    }
}

#ifdef CAN_TELEPORT
void PlayField::teleport(void)
{
    int i;

    if (m_score >= 1000) {
	/* try to teleport to the start position, or any other that is far
	 * away from flames.
	 */
	int x = 0;
	int y = 0;
	int nloops = 0;
	while (++nloops < 100) {
	    if (detectCollision(0, x, y) < 0) {
		for (i = 1; i < m_actors; i++) {
		    if (m_actor[i].m_type == Flame) {
			int xdiff = abs(m_actor[i].m_x - x);
			int ydiff = abs(m_actor[i].m_y - y);
			if (xdiff < 90 || ydiff < 90) {
			    break;
			}
		    }
		}
		if (i >= m_actors) {
		    m_score = m_score - 1000;
		    m_actor[0].m_x = x;
		    m_actor[0].m_y = y;
		    m_actor[0].m_counter = 0;
		    m_actor[0].m_req_dir = Stop;
		    m_actor[0].m_act_dir = Stop;
		    return;
		}
	    }
	    x = (rand() % max_cols) * 30;
	    y = (rand() % max_cols) * 30;
	}
    }
}
#endif

int PlayField::detectCollision(int actor, int x, int y)
{
    int i;

    /* if we would hit an outside wall, just return the square of any wall on the map */
    if (x < 0 || y < 0 || x > ((max_cols - 1) * 30) || y > ((max_rows - 1) * 30)) {
	for (i = 0; i < m_actors; i++) {
	    if (m_actor[i].m_type == Wall) {
		return i;
	    }
	}
	dump(__FILE__, __FUNCTION__, __LINE__);
    }

    /* check for the known objects */
    for (i = 0; i < m_actors; i++) {
	if (i != actor && m_actor[i].m_type != Empty && m_actor[i].m_type != Points) {
	    int xdiff = abs(m_actor[i].m_x - x);
	    int ydiff = abs(m_actor[i].m_y - y);
	    if (xdiff < 30 && ydiff < 30) {
		return i;
	    }
	}
    }

    return -1;
}

int PlayField::detectCollisionMovingBlock(int actor, int x, int y)
{
    int i;

    for (i = 0; i < m_actors; i++) {
	if (i != actor && (m_actor[i].m_type == Block || m_actor[i].m_type == BlockCoin)) {
	    int xdiff = abs(m_actor[i].m_x - x);
	    int ydiff = abs(m_actor[i].m_y - y);
	    if (xdiff < 30 && ydiff < 30 && m_actor[i].m_act_dir != Stop) {
		return i;
	    }
	}
    }
    return -1;
}

int PlayField::detectCollisionBlockOrWall(int actor, int x, int y)
{
    int i;

    /* if we would hit an outside wall, just return the square of any wall on the map */
    if (x < 0 || y < 0 || x > ((max_cols - 1) * 30) || y > ((max_rows - 1) * 30)) {
	for (i = 0; i < m_actors; i++) {
	    if (m_actor[i].m_type == Wall) {
		return i;
	    }
	}
	dump(__FILE__, __FUNCTION__, __LINE__);
    }

    /* check for the known objects */
    for (i = 0; i < m_actors; i++) {
	if (i != actor
	    && (m_actor[i].m_type == Block || m_actor[i].m_type == BlockCoin
		|| m_actor[i].m_type == Wall)) {
	    int xdiff = abs(m_actor[i].m_x - x);
	    int ydiff = abs(m_actor[i].m_y - y);
	    if (xdiff < 30 && ydiff < 30) {
		return i;
	    }
	}
    }

    return -1;
}


int PlayField::detectCollisionFlame(int actor, int x, int y)
{
    int i;

    for (i = 0; i < m_actors; i++) {
	if (i != actor && m_actor[i].m_type == Flame) {
	    int xdiff = abs(m_actor[i].m_x - x);
	    int ydiff = abs(m_actor[i].m_y - y);
	    if (xdiff < 30 && ydiff < 30) {
		return i;
	    }
	}
    }
    return -1;
}

int PlayField::detectCollisionPlayer(int x, int y)
{
    int xdiff = abs(m_actor[0].m_x - x);
    int ydiff = abs(m_actor[0].m_y - y);
    if (xdiff < 30 && ydiff < 30) {
	return 0;
    } else {
	return -1;
    }
}

int PlayField::detectCollisionWall(int actor, int x, int y)
{
    int i;

    /* if we would hit an outside wall, just return the square of any wall on the map */
    if (x < 0 || y < 0 || x > ((max_cols - 1) * 30) || y > ((max_rows - 1) * 30)) {
	for (i = 0; i < m_actors; i++) {
	    if (m_actor[i].m_type == Wall) {
		return i;
	    }
	}
	dump(__FILE__, __FUNCTION__, __LINE__);
    }

    /* check for the known objects */
    for (i = 0; i < m_actors; i++) {
	if (i != actor && m_actor[i].m_type == Wall) {
	    int xdiff = abs(m_actor[i].m_x - x);
	    int ydiff = abs(m_actor[i].m_y - y);
	    if (xdiff < 30 && ydiff < 30) {
		return i;
	    }
	}
    }

    return -1;
}

bool PlayField::hasReachedClickPosition(int x, int y)
{
    if (m_click_x != -1 && m_click_y != -1) {
	x = (x / 30) * 30;
	y = (y / 30) * 30;
	if (x == m_click_x && y == m_click_y) {
	    return true;
	}
    }
    return false;
}

bool PlayField::lineOfSight(int actor1, int actor2, enum Directions dir)
{
    int x1 = m_actor[actor1].m_x;
    int y1 = m_actor[actor1].m_y;
    int x2 = m_actor[actor2].m_x;
    int y2 = m_actor[actor2].m_y;
    int i;

    /* if we look from actor1 in direction dir to actor2, is there any
     * obstacle between actor1 and actor2?
     */
    for (i = 0; i < m_actors; i++) {
	if (i != actor1 && i != actor2) {
	    switch (dir) {
	    case Up:
		if (m_actor[i].m_y < y1 && m_actor[i].m_y > y2) {
		    int xdiff1 = abs(m_actor[i].m_x - x1);
		    int xdiff2 = abs(m_actor[i].m_x - x2);
		    if (xdiff1 < 15 || xdiff2 < 15) {
			return false;
		    }
		}
		break;
	    case Down:
		if (m_actor[i].m_y > y1 && m_actor[i].m_y < y2) {
		    int xdiff1 = abs(m_actor[i].m_x - x1);
		    int xdiff2 = abs(m_actor[i].m_x - x2);
		    if (xdiff1 < 15 || xdiff2 < 15) {
			return false;
		    }
		}
		break;
	    case Left:
		if (m_actor[i].m_x < x1 && m_actor[i].m_x > x2) {
		    int ydiff1 = abs(m_actor[i].m_y - y1);
		    int ydiff2 = abs(m_actor[i].m_y - y2);
		    if (ydiff1 < 15 || ydiff2 < 15) {
			return false;
		    }
		}
		break;
	    case Right:
		if (m_actor[i].m_x > x1 && m_actor[i].m_x < x2) {
		    int ydiff1 = abs(m_actor[i].m_y - y1);
		    int ydiff2 = abs(m_actor[i].m_y - y2);
		    if (ydiff1 < 15 || ydiff2 < 15) {
			return false;
		    }
		}
		break;
	    default:
		break;
	    }
	}
    }
    return true;
}

void PlayField::dump(const char *file, const char *function, int line)
{
    int i;

    printf("Dump in %s/%s, line %d:\n", file, function, line);
    printf("\n");
    printf("speed:%d, strength:%d, coins:%d\n", m_speed, m_strength, m_coins);
    printf("state:%d, counter:%d\n", m_state, m_counter);
    printf("score:%d, lives:%d, levels:%d, blocks:%d, coins:%d, flames:%d\n",
	m_score, m_lives, m_levels_completed, m_blocks_broken, m_coins_collected,
	m_flames_killed);
#ifdef CAN_TELEPORT
    printf("actors:%d, coinsleft:%d, click_x:%d, clicky:%d, teleport:%d\n",
	m_actors, m_coins_in_level, m_click_x, m_click_y, m_teleport);
#else
    printf("actors:%d, coinsleft:%d, click_x:%d, clicky:%d\n",
	m_actors, m_coins_in_level, m_click_x, m_click_y);
#endif
    printf("\n");
    for (i = 0; i < m_actors; i++) {
	printf("actor[%d] = {x:%d, y:%d, act_dir:%d, req_dir:%d, speed:%d, counter:%d, type:%d}\n",
	    i, m_actor[i].m_x, m_actor[i].m_y, m_actor[i].m_act_dir, m_actor[i].m_req_dir,
	    m_actor[i].m_speed, m_actor[i].m_counter, m_actor[i].m_type);
    }
    printf("\n");
}

